@@ -7,11 +7,7 @@ devices.
 This supports BlueZ 5 only.
 """
 
-# This program is free software; you can redistribute it and/or modify it under
-# the terms of the GNU Lesser General Public License as published by the Free
-# Software Foundation; either version 3 of the License, or (at your option) any
-# later version.  See http://www.gnu.org/copyleft/lgpl.html for the full text
-# of the license.
+# SPDX-License-Identifier: LGPL-3.0-or-later
 
 __author__ = "Philip Withnall"
 __copyright__ = """
@@ -38,9 +34,17 @@ MEDIA_IFACE = "org.bluez.Media1"
 NETWORK_SERVER_IFACE = "org.bluez.Network1"
 DEVICE_IFACE = "org.bluez.Device1"
 
+LE_ADVERTISING_MANAGER_IFACE = "org.bluez.LEAdvertisingManager1"
+LE_ADVERTISEMENT_IFACE = "org.bluez.LEAdvertisement1"
+ADVERTISEMENT_MONITOR_MANAGER_IFACE = "org.bluez.AdvertisementMonitorManager1"
+ADVERTISEMENT_MONITOR_IFACE = "org.bluez.AdvertisementMonitor1"
+
 # The device class of some arbitrary Android phone.
 MOCK_PHONE_CLASS = 5898764
 
+# Maximum number of BLE advertisements supported per adapter.
+MAX_ADVERTISEMENT_INSTANCES = 5
+
 
 @dbus.service.method(AGENT_MANAGER_IFACE, in_signature="os", out_signature="")
 def RegisterAgent(manager, agent_path, capability):
@@ -48,9 +52,13 @@ def RegisterAgent(manager, agent_path, capability):
 
     if agent_path in manager.agent_paths:
         raise dbus.exceptions.DBusException(
-            "Another agent is already registered " + manager.agent_path, name="org.bluez.Error.AlreadyExists"
+            "Another agent is already registered " + agent_path, name="org.bluez.Error.AlreadyExists"
         )
 
+    # Fallback to "KeyboardDisplay" as per BlueZ spec
+    if not capability:
+        capability = "KeyboardDisplay"
+
     if capability not in all_caps:
         raise dbus.exceptions.DBusException(
             "Unsupported capability " + capability, name="org.bluez.Error.InvalidArguments"
@@ -83,7 +91,7 @@ def RequestDefaultAgent(manager, agent_path):
     manager.default_agent = agent_path
 
 
-def load(mock, _parameters):
+def load(mock, parameters):
     mock.AddObject(
         "/org/bluez",
         AGENT_MANAGER_IFACE,
@@ -107,6 +115,11 @@ def load(mock, _parameters):
     bluez.capabilities = {}
     bluez.default_agent = None
 
+    # whether to expose the LEAdvertisingManager1 interface on adapters (BLE advertising)
+    bluez.enable_advertise_api = parameters.get("enable_advertise_api", True)
+    # whether to expose the AdvertisementMonitorManager1 interface on adapters (Passive scanning)
+    bluez.enable_monitor_api = parameters.get("enable_monitor_api", True)
+
 
 @dbus.service.method(ADAPTER_IFACE, in_signature="o", out_signature="")
 def RemoveDevice(adapter, path):
@@ -138,8 +151,8 @@ def StartDiscovery(adapter):
         [
             ADAPTER_IFACE,
             {
-                "Discoverable": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discoverable"], variant_level=1),
-                "Discovering": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discovering"], variant_level=1),
+                "Discoverable": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discoverable"]),
+                "Discovering": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discovering"]),
             },
             [],
         ],
@@ -160,8 +173,8 @@ def StopDiscovery(adapter):
         [
             ADAPTER_IFACE,
             {
-                "Discoverable": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discoverable"], variant_level=1),
-                "Discovering": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discovering"], variant_level=1),
+                "Discoverable": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discoverable"]),
+                "Discovering": dbus.Boolean(adapter.props[ADAPTER_IFACE]["Discovering"]),
             },
             [],
         ],
@@ -204,23 +217,23 @@ def AddAdapter(self, device_name, system_name):
                 # Audio/Video Remote Control Profile (target)
                 "0000110c-0000-1000-8000-00805f9b34fb",
             ],
-            variant_level=1,
         ),
-        "Discoverable": dbus.Boolean(False, variant_level=1),
-        "Discovering": dbus.Boolean(False, variant_level=1),
-        "Pairable": dbus.Boolean(True, variant_level=1),
-        "Powered": dbus.Boolean(True, variant_level=1),
-        "Address": dbus.String(address, variant_level=1),
-        "AddressType": dbus.String("public", variant_level=1),
-        "Alias": dbus.String(system_name, variant_level=1),
-        "Modalias": dbus.String("usb:v1D6Bp0245d050A", variant_level=1),
-        "Name": dbus.String(system_name, variant_level=1),
+        "Discoverable": dbus.Boolean(False),
+        "Discovering": dbus.Boolean(False),
+        "Pairable": dbus.Boolean(True),
+        "Powered": dbus.Boolean(True),
+        "Address": dbus.String(address),
+        "AddressType": dbus.String("public"),
+        "Alias": dbus.String(system_name),
+        "Modalias": dbus.String("usb:v1D6Bp0245d050A"),
+        "Name": dbus.String(system_name),
         # Reference:
         # http://bluetooth-pentest.narod.ru/software/
         # bluetooth_class_of_device-service_generator.html
-        "Class": dbus.UInt32(268, variant_level=1),  # Computer, Laptop
-        "DiscoverableTimeout": dbus.UInt32(180, variant_level=1),
-        "PairableTimeout": dbus.UInt32(0, variant_level=1),
+        "Class": dbus.UInt32(268),  # Computer, Laptop
+        "DiscoverableTimeout": dbus.UInt32(180),
+        "PairableTimeout": dbus.UInt32(0),
+        "Roles": dbus.Array(["central", "peripheral"]),
     }
 
     self.AddObject(
@@ -253,6 +266,61 @@ def AddAdapter(self, device_name, system_name):
         ],
     )
 
+    bluez = mockobject.objects["/org/bluez"]
+
+    # Advertising Manager
+    if bluez.enable_advertise_api:
+        # Example values below from an Intel AX200 adapter
+        advertising_manager_properties = {
+            "ActiveInstances": dbus.Byte(0),
+            "SupportedInstances": dbus.Byte(MAX_ADVERTISEMENT_INSTANCES),
+            "SupportedIncludes": dbus.Array(["tx-power", "appearance", "local-name", "rssi"]),
+            "SupportedSecondaryChannels": dbus.Array(["1M", "2M", "Coded"]),
+            "SupportedCapabilities": dbus.Dictionary(
+                {
+                    "MaxAdvLen": dbus.Byte(251),
+                    "MaxScnRspLen": dbus.Byte(251),
+                    "MinTxPower": dbus.Int16(-34),
+                    "MaxTxPower": dbus.Int16(7),
+                },
+                signature="sv",
+            ),
+            "SupportedFeatures": dbus.Array(
+                [
+                    "CanSetTxPower",
+                    "HardwareOffload",
+                ],
+            ),
+        }
+        adapter.AddProperties(LE_ADVERTISING_MANAGER_IFACE, advertising_manager_properties)
+        adapter.AddMethods(
+            LE_ADVERTISING_MANAGER_IFACE,
+            [
+                ("RegisterAdvertisement", "oa{sv}", "", RegisterAdvertisement),
+                ("UnregisterAdvertisement", "o", "", UnregisterAdvertisement),
+            ],
+        )
+
+        # Track advertisements per adapter
+        adapter.advertisements = []
+
+    # Advertisement Monitor Manager
+    if bluez.enable_monitor_api:
+        advertisement_monitor_manager_properties = {
+            "SupportedMonitorTypes": dbus.Array(["or_patterns"]),
+        }
+        adapter.AddProperties(ADVERTISEMENT_MONITOR_MANAGER_IFACE, advertisement_monitor_manager_properties)
+        adapter.AddMethods(
+            ADVERTISEMENT_MONITOR_MANAGER_IFACE,
+            [
+                ("RegisterMonitor", "o", "", RegisterMonitor),
+                ("UnregisterMonitor", "o", "", UnregisterMonitor),
+            ],
+        )
+
+        # Track advertisement monitors per adapter
+        adapter.monitors = []
+
     manager = mockobject.objects["/"]
     manager.EmitSignal(
         OBJECT_MANAGER_IFACE,
@@ -348,7 +416,7 @@ def Connect(device):
         [
             DEVICE_IFACE,
             {
-                "Connected": dbus.Boolean(device.connected, variant_level=1),
+                "Connected": dbus.Boolean(device.connected),
             },
             [],
         ],
@@ -367,7 +435,7 @@ def Disconnect(device):
         [
             DEVICE_IFACE,
             {
-                "Connected": dbus.Boolean(device.connected, variant_level=1),
+                "Connected": dbus.Boolean(device.connected),
             },
             [],
         ],
@@ -399,29 +467,29 @@ def AddDevice(self, adapter_device_name, device_address, alias):
         )
 
     properties = {
-        "Address": dbus.String(device_address, variant_level=1),
-        "AddressType": dbus.String("public", variant_level=1),
-        "Name": dbus.String(alias, variant_level=1),
-        "Icon": dbus.String("phone", variant_level=1),
-        "Class": dbus.UInt32(MOCK_PHONE_CLASS, variant_level=1),
-        "Appearance": dbus.UInt16(0, variant_level=1),
-        "UUIDs": dbus.Array([], signature="s", variant_level=1),
-        "Paired": dbus.Boolean(False, variant_level=1),
-        "Connected": dbus.Boolean(False, variant_level=1),
-        "Trusted": dbus.Boolean(False, variant_level=1),
-        "Blocked": dbus.Boolean(False, variant_level=1),
-        "WakeAllowed": dbus.Boolean(False, variant_level=1),
-        "Alias": dbus.String(alias, variant_level=1),
-        "Adapter": dbus.ObjectPath(adapter_path, variant_level=1),
-        "LegacyPairing": dbus.Boolean(False, variant_level=1),
-        "Modalias": dbus.String("bluetooth:v000Fp1200d1436", variant_level=1),
-        "RSSI": dbus.Int16(-79, variant_level=1),  # arbitrary
-        "TxPower": dbus.Int16(0, variant_level=1),
-        "ManufacturerData": dbus.Array([], signature="a{qv}", variant_level=1),
-        "ServiceData": dbus.Array([], signature="a{sv}", variant_level=1),
-        "ServicesResolved": dbus.Boolean(False, variant_level=1),
-        "AdvertisingFlags": dbus.Array([], signature="ay", variant_level=1),
-        "AdvertisingData": dbus.Array([], signature="a{yv}", variant_level=1),
+        "Address": dbus.String(device_address),
+        "AddressType": dbus.String("public"),
+        "Name": dbus.String(alias),
+        "Icon": dbus.String("phone"),
+        "Class": dbus.UInt32(MOCK_PHONE_CLASS),
+        "Appearance": dbus.UInt16(0),
+        "UUIDs": dbus.Array([], signature="s"),
+        "Paired": dbus.Boolean(False),
+        "Connected": dbus.Boolean(False),
+        "Trusted": dbus.Boolean(False),
+        "Blocked": dbus.Boolean(False),
+        "WakeAllowed": dbus.Boolean(False),
+        "Alias": dbus.String(alias),
+        "Adapter": dbus.ObjectPath(adapter_path),
+        "LegacyPairing": dbus.Boolean(False),
+        "Modalias": dbus.String("bluetooth:v000Fp1200d1436"),
+        "RSSI": dbus.Int16(-79),  # arbitrary
+        "TxPower": dbus.Int16(0),
+        "ManufacturerData": dbus.Array([], signature="a{qv}"),
+        "ServiceData": dbus.Array([], signature="a{sv}"),
+        "ServicesResolved": dbus.Boolean(False),
+        "AdvertisingFlags": dbus.Array([], signature="ay"),
+        "AdvertisingData": dbus.Array([], signature="a{yv}"),
     }
 
     self.AddObject(
@@ -504,10 +572,10 @@ def PairDevice(_self, adapter_device_name, device_address):
     device.UpdateProperties(
         DEVICE_IFACE,
         {
-            "UUIDs": dbus.Array(uuids, variant_level=1),
-            "Paired": dbus.Boolean(True, variant_level=1),
-            "LegacyPairing": dbus.Boolean(True, variant_level=1),
-            "Blocked": dbus.Boolean(False, variant_level=1),
+            "UUIDs": dbus.Array(uuids),
+            "Paired": dbus.Boolean(True),
+            "LegacyPairing": dbus.Boolean(True),
+            "Blocked": dbus.Boolean(False),
         },
     )
 
@@ -542,8 +610,8 @@ def BlockDevice(_self, adapter_device_name, device_address):
 
     device = mockobject.objects[device_path]
 
-    device.props[DEVICE_IFACE]["Blocked"] = dbus.Boolean(True, variant_level=1)
-    device.props[DEVICE_IFACE]["Connected"] = dbus.Boolean(False, variant_level=1)
+    device.props[DEVICE_IFACE]["Blocked"] = dbus.Boolean(True)
+    device.props[DEVICE_IFACE]["Connected"] = dbus.Boolean(False)
 
     device.EmitSignal(
         dbus.PROPERTIES_IFACE,
@@ -552,8 +620,8 @@ def BlockDevice(_self, adapter_device_name, device_address):
         [
             DEVICE_IFACE,
             {
-                "Blocked": dbus.Boolean(True, variant_level=1),
-                "Connected": dbus.Boolean(False, variant_level=1),
+                "Blocked": dbus.Boolean(True),
+                "Connected": dbus.Boolean(False),
             },
             [],
         ],
@@ -590,8 +658,8 @@ def ConnectDevice(_self, adapter_device_name, device_address):
 
     device = mockobject.objects[device_path]
 
-    device.props[DEVICE_IFACE]["Blocked"] = dbus.Boolean(False, variant_level=1)
-    device.props[DEVICE_IFACE]["Connected"] = dbus.Boolean(True, variant_level=1)
+    device.props[DEVICE_IFACE]["Blocked"] = dbus.Boolean(False)
+    device.props[DEVICE_IFACE]["Connected"] = dbus.Boolean(True)
 
     device.EmitSignal(
         dbus.PROPERTIES_IFACE,
@@ -600,8 +668,8 @@ def ConnectDevice(_self, adapter_device_name, device_address):
         [
             DEVICE_IFACE,
             {
-                "Blocked": dbus.Boolean(False, variant_level=1),
-                "Connected": dbus.Boolean(True, variant_level=1),
+                "Blocked": dbus.Boolean(False),
+                "Connected": dbus.Boolean(True),
             },
             [],
         ],
@@ -638,7 +706,7 @@ def DisconnectDevice(_self, adapter_device_name, device_address):
 
     device = mockobject.objects[device_path]
 
-    device.props[DEVICE_IFACE]["Connected"] = dbus.Boolean(False, variant_level=1)
+    device.props[DEVICE_IFACE]["Connected"] = dbus.Boolean(False)
 
     device.EmitSignal(
         dbus.PROPERTIES_IFACE,
@@ -647,8 +715,140 @@ def DisconnectDevice(_self, adapter_device_name, device_address):
         [
             DEVICE_IFACE,
             {
-                "Connected": dbus.Boolean(False, variant_level=1),
+                "Connected": dbus.Boolean(False),
             },
             [],
         ],
     )
+
+
+def RegisterAdvertisement(manager, adv_path, options):  # pylint: disable=unused-argument
+    if adv_path in manager.advertisements:
+        raise dbus.exceptions.DBusException("Already registered: " + adv_path, name="org.bluez.Error.AlreadyExists")
+
+    if len(manager.advertisements) >= MAX_ADVERTISEMENT_INSTANCES:
+        raise dbus.exceptions.DBusException(
+            f"Maximum number of advertisements reached: {MAX_ADVERTISEMENT_INSTANCES}",
+            name="org.bluez.Error.NotPermitted",
+        )
+
+    manager.advertisements.append(adv_path)
+
+    manager.UpdateProperties(
+        LE_ADVERTISING_MANAGER_IFACE,
+        {
+            "ActiveInstances": dbus.Byte(len(manager.advertisements)),
+            "SupportedInstances": dbus.Byte(MAX_ADVERTISEMENT_INSTANCES - len(manager.advertisements)),
+        },
+    )
+
+
+def UnregisterAdvertisement(manager, adv_path):
+    try:
+        manager.advertisements.remove(adv_path)
+    except ValueError:
+        raise dbus.exceptions.DBusException(
+            "Unknown advertisement: " + adv_path, name="org.bluez.Error.DoesNotExist"
+        ) from None
+
+    manager.UpdateProperties(
+        LE_ADVERTISING_MANAGER_IFACE,
+        {
+            "ActiveInstances": dbus.Byte(len(manager.advertisements)),
+            "SupportedInstances": dbus.Byte(MAX_ADVERTISEMENT_INSTANCES - len(manager.advertisements)),
+        },
+    )
+
+
+@dbus.service.method(BLUEZ_MOCK_IFACE, in_signature="s", out_signature="s")
+def AddAdvertisement(self, adv_name):
+    """Convenience method to add an Advertisement object
+
+    Creates a simple broadcast advertisement with some manufacturer data.
+
+    Returns the new object path.
+    """
+    path = "/org/dbusmock/bluez/advertisement/" + adv_name
+
+    adv_properties = {
+        "Type": dbus.String("broadcast"),
+        "ManufacturerData": dbus.Dictionary(
+            # 0xFFFF is the Bluetooth Company Identifier reserved for internal use and testing.
+            {dbus.UInt16(0xFFFF): dbus.Array([0x00, 0x01])},
+            signature="qv",
+        ),
+        "Includes": dbus.Array(["local-name"]),
+    }
+
+    self.AddObject(
+        path,
+        LE_ADVERTISEMENT_IFACE,
+        adv_properties,
+        [
+            ("Release", "", "", ""),
+        ],
+    )
+
+    return path
+
+
+@dbus.service.method(BLUEZ_MOCK_IFACE, in_signature="s", out_signature="s")
+def AddMonitor(self, monitor_name):
+    """Convenience method to add an Advertisement Monitor
+
+    Returns the new object path.
+    """
+    path = "/org/dbusmock/bluez/monitor/" + monitor_name
+
+    monitor_properties = {
+        "Type": dbus.String("or_patterns"),
+        # Example pattern that could be used to scan for an advertisement created by AddAdvertisement()
+        "Patterns": dbus.Struct(
+            (
+                # Start position: 0
+                dbus.Byte(0),
+                # AD data type: Manufacturer data
+                dbus.Byte(0xFF),
+                # Vaue of the pattern: 0xFFFF (company identifier), followed by 0x01
+                dbus.Array(
+                    [
+                        dbus.UInt16(0xFFFF),
+                        dbus.Byte(0x01),
+                    ]
+                ),
+            ),
+            signature="yyay",
+        ),
+    }
+
+    self.AddObject(
+        path,
+        ADVERTISEMENT_MONITOR_IFACE,
+        monitor_properties,
+        [
+            ("Release", "", "", ""),
+            ("Activate", "", "", ""),
+            ("DeviceFound", "o", "", ""),
+            ("DeviceLost", "o", "", ""),
+        ],
+    )
+
+    return path
+
+
+def RegisterMonitor(manager, monitor_path):
+    if monitor_path in manager.monitors:
+        raise dbus.exceptions.DBusException(
+            "Already registered: " + monitor_path, name="org.bluez.Error.AlreadyExists"
+        )
+
+    manager.monitors.append(monitor_path)
+
+
+def UnregisterMonitor(manager, monitor_path):
+    try:
+        manager.monitors.remove(monitor_path)
+    except ValueError:
+        raise dbus.exceptions.DBusException(
+            "Unknown monitor: " + monitor_path, name="org.bluez.Error.DoesNotExist"
+        ) from None
